with GNATCOLL.Scripts;

package Support is

   procedure Register_Functions (Repo : GNATCOLL.Scripts.Scripts_Repository);
   --  Register the various scripting languages and the functions we export
   --  to them

end Support;
