XML = r'''<?xml version="1.0"?>
<GPS>
  <documentation_file>
     <shell>Editor.edit "gnatcoll-format_columns_vertical.ads"</shell>
     <descr></descr>
     <menu>/Help/GNAT Runtime/</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-format_columns_vertical_xstring.ads"</shell>
     <descr></descr>
     <menu>/Help/GNAT Runtime/</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-any_types.ads"</shell>
     <descr>GNATCOLL.Any_Types</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Any__Types/&lt;Any__Types&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-any_types-python.ads"</shell>
     <descr>GNATCOLL.Any_Types.Python</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Any__Types/Python</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-arg_lists.ads"</shell>
     <descr>GNATCOLL.Arg_Lists</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Arg__Lists</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-asserts.ads"</shell>
     <descr>GNATCOLL.Asserts</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Asserts</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-atomic.ads"</shell>
     <descr>GNATCOLL.Atomic</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Atomic</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-boyer_moore.ads"</shell>
     <descr>GNATCOLL.Boyer_Moore</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Boyer__Moore</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-config.ads"</shell>
     <descr>GNATCOLL.Config</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Config</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-email.ads"</shell>
     <descr>GNATCOLL.Email</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Email/&lt;Email&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-email-mailboxes.ads"</shell>
     <descr>GNATCOLL.Email.Mailboxes</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Email/Mailboxes</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-email-parser.ads"</shell>
     <descr>GNATCOLL.Email.Parser</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Email/Parser</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-email-utils.ads"</shell>
     <descr>GNATCOLL.Email.Utils</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Email/Utils</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-formatters.ads"</shell>
     <descr>GNATCOLL.Formatters</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Formatters</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp.ads"</shell>
     <descr>GNATCOLL.GMP</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/&lt;GMP&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp-integers.ads"</shell>
     <descr>GNATCOLL.GMP.Integers</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/Integers/&lt;Integers&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp-integers-io.ads"</shell>
     <descr>GNATCOLL.GMP.Integers.IO</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/Integers/IO</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp-integers-misc.ads"</shell>
     <descr>GNATCOLL.GMP.Integers.Misc</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/Integers/Misc</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp-integers-number_theoretic.ads"</shell>
     <descr>GNATCOLL.GMP.Integers.Number_Theoretic</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/Integers/Number__Theoretic</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp-integers-random.ads"</shell>
     <descr>GNATCOLL.GMP.Integers.Random</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/Integers/Random</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp-integers-root_extraction.ads"</shell>
     <descr>GNATCOLL.GMP.Integers.Root_Extraction</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/Integers/Root__Extraction</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp-lib.ads"</shell>
     <descr>GNATCOLL.GMP.Lib</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/Lib</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-gmp-random_state.ads"</shell>
     <descr>GNATCOLL.GMP.Random_State</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/GMP/Random__State</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-geometry.ads"</shell>
     <descr>GNATCOLL.Geometry</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Geometry</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-io-native.ads"</shell>
     <descr>GNATCOLL.IO.Native</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/IO/Native</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-io-remote.ads"</shell>
     <descr>GNATCOLL.IO.Remote</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/IO/Remote/&lt;Remote&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-io-remote-unix.ads"</shell>
     <descr>GNATCOLL.IO.Remote.Unix</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/IO/Remote/Unix</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-io-remote-windows.ads"</shell>
     <descr>GNATCOLL.IO.Remote.Windows</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/IO/Remote/Windows</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-iconv.ads"</shell>
     <descr>GNATCOLL.Iconv</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Iconv</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-json.ads"</shell>
     <descr>GNATCOLL.JSON</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/JSON</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-memory.ads"</shell>
     <descr>GNATCOLL.Memory</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Memory</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-mmap.ads"</shell>
     <descr>GNATCOLL.Mmap</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Mmap/&lt;Mmap&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-paragraph_filling.ads"</shell>
     <descr>GNATCOLL.Paragraph_Filling</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Paragraph__Filling</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-pools.ads"</shell>
     <descr>GNATCOLL.Pools</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Pools</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-projects.ads"</shell>
     <descr>GNATCOLL.Projects</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Projects/&lt;Projects&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-projects-aux.ads"</shell>
     <descr>GNATCOLL.Projects.Aux</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Projects/Aux</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-promises.ads"</shell>
     <descr>GNATCOLL.Promises</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Promises</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-python.ads"</shell>
     <descr>GNATCOLL.Python</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Python</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar.ads"</shell>
     <descr>GNATCOLL.Ravenscar</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/&lt;Ravenscar&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-multiple_queue_cyclic_server.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Multiple_Queue_Cyclic_Server</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Multiple__Queue__Cyclic__Server</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-multiple_queue_sporadic_server.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Multiple_Queue_Sporadic_Server</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Multiple__Queue__Sporadic__Server</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-simple_cyclic_task.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Simple_Cyclic_Task</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Simple__Cyclic__Task</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-simple_sporadic_task.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Simple_Sporadic_Task</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Simple__Sporadic__Task</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-sporadic_server.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Sporadic_Server</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Sporadic__Server</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-sporadic_server_with_callback.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Sporadic_Server_With_Callback</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Sporadic__Server__With__Callback</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-timed_out_sporadic_server.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Timed_Out_Sporadic_Server</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Timed__Out__Sporadic__Server</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-timers.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Timers</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Timers/&lt;Timers&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-timers-one_shot_timer.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Timers.One_Shot_Timer</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Timers/One__Shot__Timer</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-ravenscar-utils.ads"</shell>
     <descr>GNATCOLL.Ravenscar.Utils</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Ravenscar/Utils</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-readline.ads"</shell>
     <descr>GNATCOLL.Readline</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Readline</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-refcount.ads"</shell>
     <descr>GNATCOLL.Refcount</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Refcount/&lt;Refcount&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-refcount-weakref.ads"</shell>
     <descr>GNATCOLL.Refcount.Weakref</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Refcount/Weakref</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-remote.ads"</shell>
     <descr>GNATCOLL.Remote</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Remote/&lt;Remote&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-remote-db.ads"</shell>
     <descr>GNATCOLL.Remote.Db</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Remote/Db</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql.ads"</shell>
     <descr>GNATCOLL.SQL</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/&lt;SQL&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-exec.ads"</shell>
     <descr>GNATCOLL.SQL.Exec</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Exec/&lt;Exec&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-exec-tasking.ads"</shell>
     <descr>GNATCOLL.SQL.Exec.Tasking</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Exec/Tasking</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-inspect.ads"</shell>
     <descr>GNATCOLL.SQL.Inspect</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Inspect</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-orm.ads"</shell>
     <descr>GNATCOLL.SQL.Orm</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Orm/&lt;Orm&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-orm-impl.ads"</shell>
     <descr>GNATCOLL.SQL.Orm.Impl</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Orm/Impl</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-postgres.ads"</shell>
     <descr>GNATCOLL.SQL.Postgres</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Postgres/&lt;Postgres&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-postgres-gnade.ads"</shell>
     <descr>GNATCOLL.SQL.Postgres.Gnade</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Postgres/Gnade</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-ranges.ads"</shell>
     <descr>GNATCOLL.SQL.Ranges</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Ranges</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-sessions.ads"</shell>
     <descr>GNATCOLL.SQL.Sessions</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Sessions</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-sqlite.ads"</shell>
     <descr>GNATCOLL.SQL.Sqlite</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Sqlite/&lt;Sqlite&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql-sqlite-gnade.ads"</shell>
     <descr>GNATCOLL.SQL.Sqlite.Gnade</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL/Sqlite/Gnade</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql_fields.ads"</shell>
     <descr>GNATCOLL.SQL_Fields</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL__Fields</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-sql_impl.ads"</shell>
     <descr>GNATCOLL.SQL_Impl</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/SQL__Impl</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-scripts.ads"</shell>
     <descr>GNATCOLL.Scripts</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Scripts/&lt;Scripts&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-scripts-files.ads"</shell>
     <descr>GNATCOLL.Scripts.Files</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Scripts/Files</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-scripts-impl.ads"</shell>
     <descr>GNATCOLL.Scripts.Impl</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Scripts/Impl</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-scripts-projects.ads"</shell>
     <descr>GNATCOLL.Scripts.Projects</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Scripts/Projects</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-scripts-python.ads"</shell>
     <descr>GNATCOLL.Scripts.Python</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Scripts/Python</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-scripts-shell.ads"</shell>
     <descr>GNATCOLL.Scripts.Shell</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Scripts/Shell</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-scripts-utils.ads"</shell>
     <descr>GNATCOLL.Scripts.Utils</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Scripts/Utils</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-storage_pools.ads"</shell>
     <descr>GNATCOLL.Storage_Pools</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Storage__Pools/&lt;Storage__Pools&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-storage_pools-alignment.ads"</shell>
     <descr>GNATCOLL.Storage_Pools.Alignment</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Storage__Pools/Alignment</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-storage_pools-headers.ads"</shell>
     <descr>GNATCOLL.Storage_Pools.Headers</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Storage__Pools/Headers</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-strings.ads"</shell>
     <descr>GNATCOLL.Strings</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Strings</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-strings_impl.ads"</shell>
     <descr>GNATCOLL.Strings_Impl</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Strings__Impl</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-symbols.ads"</shell>
     <descr>GNATCOLL.Symbols</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Symbols</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-templates.ads"</shell>
     <descr>GNATCOLL.Templates</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Templates</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-terminal.ads"</shell>
     <descr>GNATCOLL.Terminal</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Terminal</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-traces.ads"</shell>
     <descr>GNATCOLL.Traces</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Traces/&lt;Traces&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-traces-syslog.ads"</shell>
     <descr>GNATCOLL.Traces.Syslog</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Traces/Syslog</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-tribooleans.ads"</shell>
     <descr>GNATCOLL.Tribooleans</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Tribooleans</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-utils.ads"</shell>
     <descr>GNATCOLL.Utils</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Utils</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-vfs.ads"</shell>
     <descr>GNATCOLL.VFS</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/VFS</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-vfs_types.ads"</shell>
     <descr>GNATCOLL.VFS_Types</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/VFS__Types</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-vfs_utils.ads"</shell>
     <descr>GNATCOLL.VFS_Utils</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/VFS__Utils</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-xref.ads"</shell>
     <descr>GNATCOLL.Xref</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Xref/&lt;Xref&gt;</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-xref-database.ads"</shell>
     <descr>GNATCOLL.Xref.Database</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Xref/Database</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

  <documentation_file>
     <shell>Editor.edit "gnatcoll-xref-database_names.ads"</shell>
     <descr>GNATCOLL.Xref.Database_Names</descr>
     <menu>/Help/GNAT Runtime/GNATCOLL/Xref/Database__Names</menu>
     <category>GNAT Components Collection</category>
  </documentation_file>

</GPS>'''
import GPS
GPS.parse_xml(XML)
